import asyncio
import logging
import typer
from typing import List
from pathlib import Path
from gidgetlab.abc import GitLabAPI
from .models import GitLabId, Project
from . import api, producers

logger = logging.getLogger(__name__)


async def run_commit_file(
    gl: GitLabAPI,
    group: GitLabId,
    local_file_path: Path,
    repo_file_path: str,
    branch: str,
    commit_message: str,
    nb_consumers: int = 5,
) -> List[Project]:
    """Start producer and consumers to commit the file to all projects from group

    Return the list of projects
    """
    content = local_file_path.read_text()
    queue: asyncio.Queue[Project] = asyncio.Queue(maxsize=0)
    consumers = [
        asyncio.create_task(
            update_projects(queue, gl, repo_file_path, branch, content, commit_message)
        )
        for _ in range(nb_consumers)
    ]
    projects = await producers.produce_projects(gl, queue, group, archived=False)
    await queue.join()
    for consumer in consumers:
        consumer.cancel()
    await gl._client.aclose()
    return projects


async def update_projects(
    queue: asyncio.Queue,
    gl: GitLabAPI,
    file_path: str,
    branch: str,
    content: str,
    commit_message: str,
) -> None:
    """Create or update the file in projects from the queue"""
    while True:
        project = await queue.get()
        typer.secho(
            f"Checking {project.path_with_namespace} (project id: {project.id})",
            fg=typer.colors.GREEN,
        )
        try:
            await api.commit_file_in_repository(
                gl,
                project.id,
                file_path,
                branch=branch,
                start_branch=branch,
                content=content,
                commit_message=commit_message,
            )
        except Exception as e:
            typer.secho(
                f"Updating {project.path_with_namespace} (project id: {project.id}) failed! {e}",
                fg=typer.colors.RED,
            )
        finally:
            queue.task_done()



import asyncio
import shutil
import typer
from typing import List
from pathlib import Path
from gidgetlab.abc import GitLabAPI
from .models import GitLabId, Project
from .util import run_subprocess
from . import producers


async def run_clone(
    gl: GitLabAPI,
    group: GitLabId,
    path: Path,
    archived: bool = False,
    delete_extra_repos: bool = False,
    force: bool = False,
    nb_consumers: int = 5,
) -> List[Project]:
    """Start producer and consumers to clone all projects from group

    Return the list of projects
    """
    queue: asyncio.Queue[Project] = asyncio.Queue(maxsize=0)
    consumers = [
        asyncio.create_task(clone_projects(queue, path, force))
        for _ in range(nb_consumers)
    ]
    projects = await producers.produce_projects(gl, queue, group, archived)
    await queue.join()
    for consumer in consumers:
        consumer.cancel()
    await gl._client.aclose()
    if delete_extra_repos:
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, delete_extra_repositories, projects, path)
    return projects


def delete_extra_repositories(projects: List[Project], path: Path) -> List[Path]:
    """Delete local git repositories not part of the list of projects

    WARNING! All projects are expected to be part of the same base namespace

    Return the list of deleted path
    """
    if not projects:
        return []
    root = path / projects[0].base_namespace
    cloned_repos = {path / project.path_with_namespace for project in projects}
    local_repos = {git_folder.parent for git_folder in root.glob("**/.git")}
    extra_repos = local_repos - cloned_repos
    for repo in extra_repos:
        typer.secho(f"Delete {repo}", fg=typer.colors.RED)
        shutil.rmtree(repo)
    return sorted(list(extra_repos))


async def clone_project(project: Project, path: Path, force: bool = False) -> None:
    """Clone or update project"""
    project_path = path / project.path_with_namespace
    if project_path.exists():
        if project.default_branch is None:
            # Git repo is empty
            typer.secho(f"Skip empty project {project.name}", fg=typer.colors.GREEN)
            return
        typer.secho(
            f"Update {project.name} under {project_path}", fg=typer.colors.GREEN
        )
        if force:
            await run_subprocess("git fetch", cwd=project_path)
            await run_subprocess(
                f"git reset --hard origin/{project.default_branch}",
                cwd=project_path,
            )
        else:
            await run_subprocess("git pull --rebase", cwd=project_path)
    else:
        parent = project_path.parent
        if not parent.exists():
            typer.secho(f"Create {parent} directory", fg=typer.colors.GREEN)
            parent.mkdir(parents=True, exist_ok=True)
        typer.secho(f"Clone {project.name} under {project_path}", fg=typer.colors.GREEN)
        await run_subprocess(f"git clone {project.ssh_url} {project_path}")


async def clone_projects(queue: asyncio.Queue, path: Path, force: bool = False) -> None:
    """Clone or update projects from queue"""
    while True:
        project = await queue.get()
        try:
            await clone_project(project, path, force)
        except Exception as e:
            typer.secho(
                f"Cloning {project.path_with_namespace} (project id: {project.id}) failed! {e}",
                fg=typer.colors.RED,
            )
        finally:
            queue.task_done()



from pydantic import BaseModel, HttpUrl
from typing import Optional, Union


GitLabId = Union[int, str]


# "full_path": "my-group",
# "id": 214,
# "kind": "group",
# "name": "my-group",
# "parent_id": null,
# "path": "my-group",
# "web_url": "https://gitlab.com/groups/my-group"
class Namespace(BaseModel):
    """GitLab Namespace"""

    id: int = -1
    name: str
    kind: str = ""
    web_url: Optional[HttpUrl] = None
    path: str = ""
    full_path: str = ""
    parent_id: Optional[int] = None

    def __str__(self) -> str:
        return self.name


# "id": 2701,
# "name": "my-cool-project",
# "description": "",
# "web_url": "https://gitlab.com/test-group/subgroup-1/my-cool-project",
# "git_ssh_url": "git@gitlab.com:test-group/subgroup-1/my-cool-project.git",
# "git_http_url": "https://gitlab.com/test-group/subgroup-1/my-cool-project.git",
# "namespace": "subgroup-1",
# "visibility_level": 0,
# "path_with_namespace": "test-group/subgroup-1/my-cool-project",
# "default_branch": "master",
class Project(BaseModel):
    """GitLab Project"""

    # Some fields are set to a default value
    # to easily create a Project manually (not from
    # a webhook or API call)
    id: int = -1
    name: str
    # description can be None
    description: Optional[str] = ""
    web_url: Optional[HttpUrl] = None
    # webhook events return git_ssh_url / git_http_url
    # as well as ssh_url / http_url
    git_ssh_url: str = ""
    git_http_url: Optional[HttpUrl] = None
    # /projects API returns ssh_url_to_repo / http_url_to_repo
    ssh_url_to_repo: str = ""
    http_url_to_repo: Optional[HttpUrl] = None
    # webhook events return a string
    # /projects API returns a Namespace
    namespace: Union[str, Namespace]
    visibility_level: int = 0
    path_with_namespace: str = ""
    # default_branch is None when the project is empty
    default_branch: Optional[str] = "master"

    class Config:
        extra = "allow"

    @property
    def base_namespace(self) -> str:
        if "/" not in self.path_with_namespace:
            return str(self.namespace)
        return self.path_with_namespace.split("/", 1)[0]

    @property
    def ssh_url(self) -> str:
        return self.ssh_url_to_repo or self.git_ssh_url

    @property
    def http_url(self) -> Optional[HttpUrl]:
        return self.http_url_to_repo or self.git_http_url

    def __str__(self) -> str:
        return self.name


class Group(BaseModel):
    """GitLab Group"""

    id: int = -1
    name: str
    web_url: Optional[HttpUrl] = None
    path: str
    description: str = ""
    full_name: str
    full_path: str
    parent_id: Optional[int] = None

    def __str__(self) -> str:
        return self.name


class Tag(BaseModel):
    """Project Tag"""

    name: str
    target: str


class TreeObject(BaseModel):
    """Tree Object"""

    id: str
    name: str
    type: str
    path: str
    mode: str



import datetime
import random
import typer
from operator import itemgetter
from typing import List
from dateutil.parser import parse
from dateutil.tz import UTC
from gidgetlab.abc import GitLabAPI
from .models import GitLabId, Project
from . import api


async def get_project_latest_pipeline_date(
    gl: GitLabAPI, project_id: GitLabId
) -> datetime.datetime:
    """Return the project latest pipeline execution date"""
    # pipelines are returned sorted by desc id by default
    # Highest id is the latest
    pipelines = await gl.getitem(f"/projects/{project_id}/pipelines")
    try:
        latest_pipeline = pipelines[0]
    except IndexError:
        return datetime.datetime(datetime.MINYEAR, 1, 1, tzinfo=UTC)
    return parse(latest_pipeline["updated_at"])


async def sort_projects_by_pipeline_date(
    gl: GitLabAPI, projects: List[Project]
) -> List[Project]:
    """Return the projects sorted by latest pipeline execution date"""
    # Build a list of tuple with each project and the latest pipeline date
    project_pipelines = [
        (project, await get_project_latest_pipeline_date(gl, project.id))
        for project in projects
    ]
    sorted_project_pipelines = sorted(project_pipelines, key=itemgetter(1))
    return [item[0] for item in sorted_project_pipelines]


async def run_trigger_group_pipelines(
    gl: GitLabAPI,
    group: GitLabId,
    nb_projects: int,
    is_random: bool = False,
    description: str = "gidgetlab",
) -> None:
    """Trigger the pipeline for the group projects

    When greater than 0, nb_projects allows to select a subset of the projects.
    If is_random is true, projects are picked randomly. If not, the ones with
    the oldest pipeline execution are selected.
    """
    # Filter out empty project (default_branch is None)
    projects = [
        project
        async for project in api.get_all_projects(gl, group)
        if project.default_branch is not None
    ]
    if nb_projects == 0 or nb_projects >= len(projects):
        sampled = projects
    else:
        if is_random:
            sampled = random.sample(projects, nb_projects)
        else:
            sorted_projects = await sort_projects_by_pipeline_date(gl, projects)
            sampled = sorted_projects[:nb_projects]
    typer.secho(
        f"Trigger pipeline for {' '.join([project.name for project in sampled])}",
        fg=typer.colors.GREEN,
    )
    await api.trigger_pipelines(gl, sampled, description)
    await gl._client.aclose()



import asyncio
from typing import List
from gidgetlab.abc import GitLabAPI
from .models import GitLabId, Project
from . import api


async def produce_projects(
    gl: GitLabAPI, queue: asyncio.Queue, group_id: GitLabId, archived: bool
) -> List[Project]:
    """Send to the queue all projects from the group, including subgroups

    Return the full list of projects when done
    """
    projects = []
    async for project in api.get_all_projects(gl, group_id, archived):
        projects.append(project)
        await queue.put(project)
    return projects



from typing import Dict, List
import asyncio
import shlex
import typer


async def run_subprocess(cmd: str, cwd=None) -> str:
    """Run the command as subprocess"""
    typer.secho(f"Run {cmd}", fg=typer.colors.CYAN)
    args = shlex.split(cmd)
    process = await asyncio.create_subprocess_exec(
        *args, stdout=asyncio.subprocess.PIPE, cwd=cwd
    )
    stdout, stderr = await process.communicate()
    return stdout.decode().strip()


def convert_params_to_dict(params: List[str]) -> Dict:
    """Convert a list of key-value pairs to a dict"""
    d = {}
    for kv in params:
        if "=" not in kv:
            raise ValueError(f"Key value pair '{kv}' shall be separated by '='")
        key, _, value = kv.partition("=")
        d[key] = value
    return d



from pprint import pprint
from typing import Dict
from gidgetlab.abc import GitLabAPI


async def run_get(gl: GitLabAPI, endpoint: str, params: Dict, nb_items: int) -> None:
    """Get one or several items"""
    result = await gl.getitem(endpoint, params=params)
    if not isinstance(result, list):
        pprint(result, sort_dicts=False)
    else:
        if nb_items > 0 and len(result) >= nb_items:
            for counter, item in enumerate(result, start=1):
                if counter > nb_items:
                    break
                pprint(item, sort_dicts=False)
        else:
            counter = 1
            async for item in gl.getiter(endpoint, params=params):
                if nb_items > 0 and counter > nb_items:
                    break
                pprint(item, sort_dicts=False)
                counter += 1
    await gl._client.aclose()



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()


requirements = [
    "gidgetlab",
    "httpx",
    "typer",
    "pydantic",
    "python-dateutil",
]
tests_requires = ["pytest", "pytest-cov", "pytest-asyncio", "pytest-mock"]

setuptools.setup(
    name="gidgetlab-kit",
    author="Benjamin Bertrand",
    author_email="beenje@gmail.com",
    description="Python package to interact with GitLab",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/beenje/gidgetlab-kit",
    license="MIT license",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    install_requires=requirements,
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="gitlab async",
    extras_require={"tests": tests_requires},
    python_requires=">=3.8",
    entry_points={"console_scripts": ["gidgetlab=gidgetlab_kit.command:app"]},
)



import asyncio
import logging
import typer
import httpx
from importlib.metadata import version, PackageNotFoundError  # type: ignore
from pathlib import Path
from typing import Optional, List
from gidgetlab.httpx import GitLabAPI
from . import util, api
from .models import GitLabId
from .clone import run_clone
from .get import run_get
from .pipelines import run_trigger_group_pipelines
from .commits import run_commit_file

try:
    __version__ = version("gidgetlab-kit")
except PackageNotFoundError:
    __version__ = "unknown"

app = typer.Typer()
state = {"gl": None}


def version_callback(value: bool):
    if value:
        typer.echo(f"gidgetlab-kit version: {__version__}")
        raise typer.Exit()


async def run_list_projects(gl: GitLabAPI, group_id: GitLabId, archived: bool) -> None:
    typer.echo(f"Projects under group {group_id}:")
    async for project in api.get_all_projects(gl, group_id, archived):
        typer.secho(
            f"{project.path_with_namespace} (id: {project.id} name: {project.name})",
            fg=typer.colors.GREEN,
        )
    await gl._client.aclose()


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show the current version and exit.",
    ),
    url: str = typer.Option("https://gitlab.com", envvar="GL_URL", help="GitLab URL"),
    access_token: str = typer.Option(
        "", envvar="GL_ACCESS_TOKEN", help="GitLab access token"
    ),
    verify: bool = typer.Option(
        True, help="Verify SSL cerificate or disable verification"
    ),
):
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        level=logging.INFO,
    )
    client = httpx.AsyncClient(verify=verify)
    state["gl"] = GitLabAPI(client, "gidgetlab", url=url, access_token=access_token)


@app.command()
def clone(
    group: str = typer.Argument(..., help="Name of the group to clone"),
    path: Path = typer.Option(
        Path("."),
        "--path",
        "-p",
        exists=True,
        resolve_path=True,
        help="Path where to clone the group",
    ),
    archived: bool = typer.Option(False, help="Clone archived projects"),
    delete_extra_repos: bool = typer.Option(
        False, help="Delete extra local repositories"
    ),
    force: bool = typer.Option(
        False,
        help="Force repository update when it exists (git reset --hard origin/default_branch instead of rebase)",
    ),
    nb_consumers: int = typer.Option(
        5,
        "--nb-consumers",
        "-n",
        help="Number of consumers to start to clone in parallel",
    ),
):
    """Clone or pull all projects from group (including subgroups)

    Projects are cloned/updated in parallel
    """
    asyncio.run(
        run_clone(
            state["gl"], group, path, archived, delete_extra_repos, force, nb_consumers
        )
    )


@app.command()
def get(
    endpoint: str = typer.Argument(..., help="Endpoint to get"),
    parameters: List[str] = typer.Argument(
        None, help="Optional key-value pairs to pass as json parameters"
    ),
    nb_items: int = typer.Option(
        0,
        "--nb-items",
        "-n",
        help="Number of items to retrieve. Only valid when a list is returned. Set to 0 to return all.",
    ),
):
    """Get one or several items from the given endpoint"""
    try:
        params = util.convert_params_to_dict(parameters)
    except ValueError:
        typer.echo(
            f"Invalid parameters: '{' '.join(parameters)}' should be a list of key-value pairs separated by '='."
        )
        raise typer.Exit(code=1)
    asyncio.run(run_get(state["gl"], endpoint, params, nb_items))


@app.command()
def list_projects(
    group: str = typer.Argument(..., help="Name of the group"),
    archived: bool = typer.Option(False, help="List archived projects"),
):
    """List all projects from group (including subgroups)"""
    asyncio.run(run_list_projects(state["gl"], group, archived))


@app.command()
def trigger_pipelines(
    group: str = typer.Argument(..., help="Name of the group"),
    nb_projects: int = typer.Option(
        0,
        "--nb-projects",
        "-n",
        help="Number of projects to trigger. Set to 0 to trigger all projects from the group.",
    ),
    random: bool = typer.Option(
        False,
        help="Randomly pick the projects. Select the oldest that were triggered otherwise.",
    ),
    description: str = typer.Option(
        "gidgetlab",
        help="Pipeline trigger description",
    ),
):
    """Trigger the pipeline for all or a subset of projects from group"""
    asyncio.run(
        run_trigger_group_pipelines(
            state["gl"], group, nb_projects, random, description
        )
    )


@app.command()
def commit_file(
    filepath: Path = typer.Argument(..., help="Local file to add or update"),
    group: str = typer.Option(..., help="Name of the group"),
    repo_path: Path = typer.Option(
        Path("."),
        "--repo-path",
        "-p",
        resolve_path=False,
        help="Path of the file in the repository",
    ),
    message: str = typer.Option(
        "",
        "--message",
        "-m",
        show_default=False,
        help="Commit message [default: Add/Update <filename>]",
    ),
    branch: str = typer.Option("master", "--branch", "-b", help="Branch to commit to"),
    nb_consumers: int = typer.Option(
        5,
        "--nb-consumers",
        "-n",
        help="Number of consumers to start to update in parallel",
    ),
):
    """Add or update a file to a list of projects

    Example:

      The following will add the local .gitlab-ci.yml file to all projects under the my-group group

      $ gidgetlab commit-file --group my-group .gitlab-ci.yml
    """
    repo_file_path = str(repo_path / filepath.name)
    asyncio.run(
        run_commit_file(
            state["gl"],
            group,
            filepath,
            repo_file_path,
            branch,
            message,
            nb_consumers,
        )
    )



import asyncio
import base64
import http
import urllib.parse
import logging
from typing import Optional, Any, AsyncIterator, List, Dict
from gidgetlab import GitLabException, HTTPException
from gidgetlab.abc import GitLabAPI
from .models import Project, Group, GitLabId, Tag, TreeObject


logger = logging.getLogger(__name__)


def url_encode(s: str) -> str:
    """Return a url encoded string

    . is also replaced by %2E
    """
    return urllib.parse.quote(s, safe="").replace(".", "%2E")


def encode_gitlab_id(gitlab_id: GitLabId) -> str:
    """Return a URL encoded GitLab id"""
    if isinstance(gitlab_id, int):
        return str(gitlab_id)
    return url_encode(gitlab_id)


async def create_file_in_repository(
    gl: GitLabAPI,
    project_id: GitLabId,
    file_path: str,
    branch: str,
    start_branch: str,
    content: str,
    commit_message: str = "",
) -> Any:
    """Create a file in the project"""
    logger.info(f"Create {file_path} in project {project_id} branch {branch}")
    commit_message = commit_message or f"Add {file_path}"
    result = await gl.post(
        f"/projects/{project_id}/repository/files/{url_encode(file_path)}",
        data={
            "branch": branch,
            "start_branch": start_branch,
            "content": content,
            "commit_message": commit_message,
        },
    )
    return result


async def update_file_in_repository(
    gl: GitLabAPI,
    project_id: GitLabId,
    file_path: str,
    branch: str,
    start_branch: str,
    content: str,
    commit_message: str = "",
) -> Any:
    """Update a file in the project"""
    logger.info(f"Update {file_path} in project {project_id} branch {branch}")
    commit_message = commit_message or f"Update {file_path}"
    result = await gl.put(
        f"/projects/{project_id}/repository/files/{url_encode(file_path)}",
        data={
            "branch": branch,
            "start_branch": start_branch,
            "content": content,
            "commit_message": commit_message,
        },
    )
    return result


async def commit_file_in_repository(
    gl: GitLabAPI,
    project_id: GitLabId,
    file_path: str,
    branch: str,
    start_branch: str,
    content: str,
    commit_message: str,
) -> Any:
    """Create or update the file in the repository

    If the file doesn't exist, it will be created.
    If the file exists, it's updated only if the content differs.
    """
    try:
        existing_content = await get_file_from_repository(
            gl, project_id, file_path, ref=start_branch
        )
    except HTTPException as e:
        if e.status_code != http.HTTPStatus.NOT_FOUND:
            raise
        # The file doesn't exist
        result = await create_file_in_repository(
            gl, project_id, file_path, branch, start_branch, content, commit_message
        )
    else:
        # The file already exists
        if existing_content == content:
            logger.info("File content is identical. Nothing to commit.")
            return
        result = await update_file_in_repository(
            gl, project_id, file_path, branch, start_branch, content, commit_message
        )
    return result


async def unsubscribe_from_merge_request(
    gl: GitLabAPI, project_id: GitLabId, mr_iid: int
) -> Any:
    """Unsubscribes the authenticated user from a merge request to not receive notifications"""
    result = await gl.post(
        f"/projects/{project_id}/merge_requests/{mr_iid}/unsubscribe", data={}
    )
    return result


async def accept_merge_request(
    gl: GitLabAPI, project_id: GitLabId, mr_iid: int, **kwargs: Any
) -> Any:
    result = await gl.put(
        f"/projects/{project_id}/merge_requests/{mr_iid}/merge", data=kwargs
    )
    return result


async def create_merge_request(
    gl: GitLabAPI,
    project_id: GitLabId,
    source_branch: str,
    target_branch: str,
    title: str,
    description: str = "",
    assignee_id: Optional[int] = None,
    remove_source_branch: bool = True,
    merge_on_success: bool = True,
    unsubscribe: bool = True,
) -> Any:
    logger.info(
        f"Create merge request '{title}' for project {project_id} from branch {source_branch} to {target_branch}"
    )
    data = {
        "source_branch": source_branch,
        "target_branch": target_branch,
        "title": title,
        "remove_source_branch": remove_source_branch,
    }
    if description:
        data["description"] = description
    if assignee_id is not None:
        data["assignee_id"] = assignee_id
    result = await gl.post(f"/projects/{project_id}/merge_requests", data=data)
    if unsubscribe:
        mr_iid = result["iid"]
        logger.info(f"Unsubscribe from merge request {mr_iid} for project {project_id}")
        await unsubscribe_from_merge_request(gl, project_id, mr_iid)
    if merge_on_success:
        # Wait to make sure the pipeline was created
        # The MR might be accepted directly otherwise
        await asyncio.sleep(5)
        mr_iid = result["iid"]
        logger.info(
            f"Accept merge request {mr_iid} for project {project_id} when pipeline succeeds"
        )
        await accept_merge_request(
            gl,
            project_id,
            mr_iid,
            merge_when_pipeline_succeeds=True,
            should_remove_source_branch=remove_source_branch,
        )
    return result


async def create_branch(
    gl: GitLabAPI, project_id: GitLabId, branch: str, ref: str
) -> Any:
    """Create a branch in the repository"""
    logger.info(f"Create branch {branch} in project {project_id} from {ref}")
    result = await gl.post(
        f"/projects/{project_id}/repository/branches",
        data={"branch": branch, "ref": ref},
    )
    return result


async def get_file_from_repository(
    gl: GitLabAPI, project_id: GitLabId, file_path: str, ref: str
) -> str:
    """Return the file from project as string"""
    result = await gl.getitem(
        f"/projects/{project_id}/repository/files/{url_encode(file_path)}",
        params={"ref": ref},
    )
    return base64.b64decode(result["content"]).decode("utf-8")


async def list_repository_tree(
    gl: GitLabAPI,
    project_id: GitLabId,
    ref: str,
    recursive: bool = True,
    path: Optional[str] = None,
) -> List[TreeObject]:
    params = {"ref": ref, "recursive": recursive}
    if path:
        params["path"] = path
    return [
        TreeObject(**repo_file)
        async for repo_file in gl.getiter(
            f"/projects/{project_id}/repository/tree",
            params=params,
        )
    ]


async def wait_for_merge_or_failure(
    gl: GitLabAPI, project_id: GitLabId, merge_request_iid: int, timeout: int = 600
) -> bool:
    """Wait for the merge request to be merged or the pipeline to fail"""
    logger.info(
        f"Wait for merge request {merge_request_iid} (project {project_id}) to be merged or pipeline to fail..."
    )
    mr_state = None
    pipeline_status = None
    sleep_time = 3
    max_loop = int(timeout / sleep_time)
    loop_nb = 0
    while (
        mr_state != "merged"
        and pipeline_status not in ("failed", "canceled")
        and loop_nb < max_loop
    ):
        loop_nb += 1
        await gl.sleep(sleep_time)
        result = await gl.getitem(
            f"/projects/{project_id}/merge_requests/{merge_request_iid}"
        )
        mr_state = result["state"]
        pipeline_status = result["head_pipeline"]["status"]
    logger.info(
        f"Merge request {merge_request_iid } state: {mr_state} / "
        f"Pipeline {result['head_pipeline']['id']} status: {pipeline_status}"
    )
    return mr_state == "merged"


async def update_file_via_merge_request(
    gl: GitLabAPI,
    project_id: GitLabId,
    file_path: str,
    content: str,
    source_branch: str,
    target_branch: str,
    message: str,
    assignee_id: Optional[int] = None,
    merge_on_success: bool = True,
    unsubscribe: bool = True,
    wait_on_merge: bool = False,
) -> None:
    await update_file_in_repository(
        gl, project_id, file_path, source_branch, target_branch, content, message
    )
    await asyncio.sleep(1)
    result = await create_merge_request(
        gl,
        project_id,
        source_branch,
        target_branch,
        message,
        assignee_id=assignee_id,
        merge_on_success=merge_on_success,
        unsubscribe=unsubscribe,
    )
    if wait_on_merge:
        if merge_on_success:
            await wait_for_merge_or_failure(gl, project_id, result["iid"])
        else:
            logger.warning(
                "wait_on_merge can only be set to True when merge_on_success is"
            )


async def get_all_projects(
    gl: GitLabAPI, group_id: GitLabId, archived: bool = False
) -> AsyncIterator[Project]:
    """Return all projects from a group, including subgroups"""
    async for project in gl.getiter(
        f"/groups/{encode_gitlab_id(group_id)}/projects", params={"archived": archived}
    ):
        yield Project(**project)
    async for subgroup in get_subgroups(gl, group_id):
        async for project in gl.getiter(
            f"/groups/{subgroup.id}/projects", params={"archived": archived}
        ):
            yield Project(**project)


async def get_subgroups(gl: GitLabAPI, group_id: GitLabId) -> AsyncIterator[Group]:
    """Return recursively all subgroups from a group"""
    async for subgroup in gl.getiter(f"/groups/{encode_gitlab_id(group_id)}/subgroups"):
        yield Group(**subgroup)
        async for sub in get_subgroups(gl, subgroup["id"]):
            yield sub


async def find_projects(
    gl: GitLabAPI, names: List[str], group_id: GitLabId
) -> List[Project]:
    """Find the projects based on their name under `group_id`

    The group's subgroups are also searched
    Return the list of projects found
    """
    names_set = {name.lower() for name in names}
    all_projects = [project async for project in get_all_projects(gl, group_id)]
    return [project for project in all_projects if project.name.lower() in names_set]


async def create_project_trigger(
    gl: GitLabAPI, project_id: GitLabId, description: str
) -> str:
    """Create a project trigger and return the corresponding token"""
    result = await gl.post(
        f"/projects/{project_id}/triggers", data={"description": description}
    )
    return result["token"]


async def get_project_triggers(
    gl: GitLabAPI, project_id: GitLabId
) -> List[Dict[str, str]]:
    """Return the project triggers"""
    return [trigger async for trigger in gl.getiter(f"/projects/{project_id}/triggers")]


async def get_or_create_project_trigger_token(
    gl: GitLabAPI, project_id: GitLabId, description: str
) -> str:
    """Return a trigger token for the project

    If no existing token is found for the given description, one is created
    """
    triggers = await get_project_triggers(gl, project_id)
    tokens = [
        trigger["token"]
        for trigger in triggers
        if trigger["description"] == description
    ]
    if tokens:
        return tokens[0]
    return await create_project_trigger(gl, project_id, description)


async def trigger_pipeline(
    gl: GitLabAPI, project_id: GitLabId, git_ref: str, token: str
) -> None:
    """Trigger the project pipeline"""
    await gl.post(
        f"/projects/{project_id}/trigger/pipeline",
        params={"token": token, "ref": git_ref},
        data={},
    )


async def list_branches(gl: GitLabAPI, project_id: GitLabId) -> List[str]:
    """Return the list of branches name"""
    return [
        branch["name"]
        async for branch in gl.getiter(f"/projects/{project_id}/repository/branches")
    ]


async def list_tags(gl: GitLabAPI, project_id: GitLabId) -> List[Tag]:
    """Return the list of tags sorted by updated field"""
    return [
        Tag(**data)
        async for data in gl.getiter(
            f"/projects/{project_id}/repository/tags",
            params={"order_by": "updated", "sort": "desc"},
        )
    ]


async def get_latest_tag_on_branch(
    gl: GitLabAPI, project_id: GitLabId, branch: str = "master"
) -> Optional[Tag]:
    """Return the latest updated tag of branch"""
    # Get all tags
    tags = await list_tags(gl, project_id)
    # Check that the tag is on the proper branch
    for tag in tags:
        async for ref in gl.getiter(
            f"/projects/{project_id}/repository/commits/{tag.target}/refs",
            params={"type": "branch"},
        ):
            if ref["type"] == "branch" and ref["name"] == branch:
                return tag
    return None


async def trigger_pipelines(
    gl: GitLabAPI,
    projects: List[Project],
    description: str,
    git_refs: Optional[List[str]] = None,
) -> None:
    """Trigger the pipeline for all projects

    If no git ref is given, the pipeline is triggered for the default branch only
    """
    for project in projects:
        if project.default_branch is None:
            # Project is empty - skip it
            continue
        try:
            token = await get_or_create_project_trigger_token(
                gl, project.id, description
            )
        except GitLabException as e:
            logger.error(f"Couldn't get project {project.name} trigger token: {e}")
            continue
        if git_refs is None:
            project_git_refs = [project.default_branch]
        else:
            project_git_refs = git_refs
        for git_ref in project_git_refs:
            try:
                await trigger_pipeline(gl, project.id, git_ref, token)
            except GitLabException as e:
                logger.error(
                    f"Couldn't trigger project {project.name} pipeline for {git_ref}: {e}"
                )
